% Przyklad kodowania Alamoutiego,
% modulacja BPSK, dwie anteny nadawcze, jedna odbiorcza

clear; close all
randn('seed',0);
rand('seed',0);

N = 10^6;          % liczba transmitowanych symboli
Eb_N0_dB = [0:25]; % Eb/N0

% kanal z zanikami Rayleigh'a
h1 = 1/sqrt(2)*[randn(1,N/2) + j*randn(1,N/2)];
h2 = 1/sqrt(2)*[randn(1,N/2) + j*randn(1,N/2)];

% wektor szumu AWGN o wariancji 1 (0dB)
n1 = 1/sqrt(2)*[randn(1,N) + j*randn(1,N)];

% strumienia danych (bit stream)
bs = rand(1,N)>0.5;

% modulacja BPSK 0 -> -1; 1 -> 0
x = 2*bs-1;

% kodowanie Alamoutiego: antena 1: [x1, x2] -> [x1, x2*]
a1 = zeros(1,N);
a1(1:2:end) = x(1:2:end);
a1(2:2:end) = conj(x(2:2:end));

% kodowanie Alamoutiego: antena 2: [x1, x2] -> [x2, -x1*]
a2 = zeros(1,N);
a2(1:2:end) = x(2:2:end);
a2(2:2:end) = -conj(x(1:2:end));

% sygnal odbierany
r1 = a1/sqrt(2) .* kron(h1, [1 1]) + a2/sqrt(2) .* kron(h2, [1 1]);

% wyznaczenie BER dla roznych poziomow szumu
for i = 1:length(Eb_N0_dB)
    % sygnal odbierany zaszumiony
    y = r1 + 10^(-Eb_N0_dB(i)/20) * n1;
    
    % rozdzielenie na dwie szczeliny
    y1 = sqrt(2) * y(1:2:end);
    y2 = sqrt(2) * y(2:2:end);
    
    % dekoder Alamoutiego(2,1)
    H2 = abs(h1).^2 + abs(h2).^2;
    x1_hat = conj(h1) .* y1 - h2 .* conj(y2);
    x2_hat = conj(h2) .* y1 + h1 .* conj(y2);
    
    % utworz jeden strumien
    x_hat(1:2:N) = x1_hat./H2;
    x_hat(2:2:N) = x2_hat./H2;
    
    % dekoder BPSK
    bs_hat = real(x_hat) > 0;
    
    % liczba blednie zdekodowanych symboli
    nErr_A21(i) = sum(bs ~= bs_hat);
end

% wyznacza BER
Ber_Alamouti21 = nErr_A21/N;


% BER dla kanalu Rayleigh i modulacji BPSK
if(0)
    % BER ze wzoru teoretycznego
    Eb_N0_Lin = 10.^(Eb_N0_dB/10);
    Ber_SISO = 1/2*(1-1./sqrt(1+1./Eb_N0_Lin));
else
    % BER wyznaczony symulacyjnie
    h = 1/sqrt(2)*[randn(1,N) + j*randn(1,N)]; %kanal
    % sygnal odbierany
    r = h .* x;
    for i = 1:length(Eb_N0_dB)
        % sygnal odbierany zaszumiony
        y = r + 10^(-Eb_N0_dB(i)/20) * n1;
        % dekoder BPSK
        x_hat = y./h;
        bs_hat = real(x_hat) > 0;
        
        % liczba blednie zdekodowanych symboli
        nErr_SISO(i) = sum(bs ~= bs_hat);
    end
    Ber_SISO = nErr_SISO/N;
end


% BER dla dekodera MRC, kanal Rayleigh, modulacja BPSK
if(0)
    % BER ze wzoru teoretycznego
    Eb_N0_Lin = 10.^(Eb_N0_dB/10);
    p = 1/2 - 1/2*(1+1./Eb_N0_Lin).^(-1/2);
    Ber_MRC_nRx2 = p.^2.*(1+2*(1-p));
else
    % BER wyznaczony symulacyjnie
    n2 = 1/sqrt(2)*[randn(1,N) + j*randn(1,N)]; %szum AWGN
    
    h1 = 1/sqrt(2)*[randn(1,N) + j*randn(1,N)]; %kanal Tx->Rx1
    h2 = 1/sqrt(2)*[randn(1,N) + j*randn(1,N)]; %kanal Tx->Rx2
    
    r1 = h1 .* x;	% sygnal odbierany przez antene nr 1
    r2 = h2 .* x;	% sygnal odbierany przez antene nr 2
    for i = 1:length(Eb_N0_dB)
        % sygnal odbierany zaszumiony
        y1 = r1 + 10^(-Eb_N0_dB(i)/20) * n1;
        y2 = r2 + 10^(-Eb_N0_dB(i)/20) * n2;
        
        % dekoder MRC
        x_hat = (y1 .* conj(h1) + y2 .* conj(h2)) ./ (h1.*conj(h1) + h2.*conj(h2));
        
        % dekoder BPSK
        bs_hat = real(x_hat) > 0;
        
        % liczba blednie zdekodowanych symboli
        nErr_MRC_nRx2(i) = sum(bs ~= bs_hat);
    end
    Ber_MRC_nRx2 = nErr_MRC_nRx2/N;
end


figure;
title('Modulacja BPSK, kanal Rayleigha');
xlabel('Eb/No, dB'); ylabel('BER');

semilogy(Eb_N0_dB, Ber_SISO, 'bp-',  ...
         Eb_N0_dB, Ber_Alamouti21,'mo-', ...
         Eb_N0_dB, Ber_MRC_nRx2,'kd-', 'LineWidth',2);

axis([0 25 10^-5 0.5]); grid on

legend( 'SISO (nTx=1, nRx=1)', ...
        'Alamouti (nTx=2, nRx=1)', ...
        'MRC (nTx=1, nRx=2)');
