% Przyklad implementacji dekoderow ML i MMSE w systemie MIMO (2, 2)

clear; close all
randn('seed',0);
rand('seed',0);

N = 10^6; % liczba transmitowanych symboli
Eb_N0_dB = [0:25]; % Eb/N0 [dB]
nTx = 2;
nRx = 2;

fprintf('start symulacji\n');

% probki czasowe realizacji kanalu MIMO Rayleigha (nRx, nTx)
h = 1/sqrt(2)*[randn(nRx, nTx, N/nTx) + j*randn(nRx, nTx, N/nTx)];

% szum AWGN o wariancji 1 (0dB)
n = 1/sqrt(2)*[randn(nRx, N/nTx) + j*randn(nRx, N/nTx)];

% generacja strumienia danych (bit stream)
bs = rand(1,N)>0.5;

% modulacja BPSK 0 -> -1; 1 -> 0
s = 2*bs-1;

% przeksztalcenie sygnalu s w macierzy o wymiarach [nRx, nTx, N/NTx]
% celem transmisji przez kanal MIMO h
% najpierw powielenie nRx razy
x = repmat(s, nRx, 1);

% potem, rozdzialenie na NTx anten nadawczych
x = reshape(x, [nRx, nTx, N/nTx]);

% transmisja przez kanal h
r = sum(h.*x, 2);
r = squeeze(r);

for i = 1:length(Eb_N0_dB)
    fprintf('------- Eb/N0 = %d [dB]\n', Eb_N0_dB(i));
    % sygnal odbierany zaszumiony
    y = r + 10^(-Eb_N0_dB(i)/20)*n;
    
    %%dekoder ML
    rVec = [];
    
    % wszystkie mozliwe kombinacje bitow dla dwoch anten nadawczych i modulacji BPSK
    x_ref = [1  1; ...
             1  0; ...
             0  1;  ...
             0  0 ];
    
    % wyznacz odleglosci dla wszystkich kombinacji z x_ref
    for k=1:size(x_ref, 1)
        % modulacja BPSK 0 -> -1; 1 -> 0
        s = 2*x_ref(k,:)-1;
        x_hat = repmat(s, [nRx ,N/2]);
        x_hat = reshape(x_hat, [nRx,nTx,N/nTx]);
        z_hat = squeeze(sum(h.*x_hat, 2)) ;
        J = sum(abs(y - z_hat), 1);
        rVec = [rVec; J];
    end
    
    % znajdz kombinacje symboli dajaca najmniejszy blad
    [jj dd] = min(rVec,[],1);
    
    % konwertuj symbole na bity
    bs_hat(1:2:N) = x_ref(dd, 1);
    bs_hat(2:2:N) = x_ref(dd, 2);
    
    % wyznacz BER
    nErr_ML(i) = sum(bs ~= bs_hat);
    
    fprintf('dekoder ML:   BER= %.3d\n', nErr_ML(i)/N);
    
    %%dekoder MMSE
    
    % Utworzenie macierzy hInv = inv(h' * h + N_0^2*I)
    % najpierw budujemy: [a b; c d] = h' * h + N_0^2*I
    N0 = 10^(-Eb_N0_dB(i)/20);
    a = sum( conj(h(:,1,:)) .* h(:,1,:), 1) + N0^2;
    b = sum( conj(h(:,1,:)) .* h(:,2,:), 1);
    c = sum( conj(h(:,2,:)) .* h(:,1,:), 1);
    d = sum( conj(h(:,2,:)) .* h(:,2,:), 1) + N0^2;
    
    % nastepnie: hInv=inv([a b; c d]) = 1/(ad-bc)[d -b;-c a]
    hInv = zeros(2,2,N/nTx);
    D = a.*d - b.*c; % ad-bc
    
    hInv(1,1,:) =  d./D;
    hInv(1,2,:) = -b./D;
    hInv(2,1,:) = -c./D;
    hInv(2,2,:) =  a./D;
    
    % Utworzenie macierzy W = hInv * h'
    W = zeros(2,2,N/nTx);
    W(1,1,:) = sum( hInv(1,:,:) .* conj(h(1,:,:)), 2);
    W(1,2,:) = sum( hInv(1,:,:) .* conj(h(2,:,:)), 2);
    W(2,1,:) = sum( hInv(2,:,:) .* conj(h(1,:,:)), 2);
    W(2,2,:) = sum( hInv(2,:,:) .* conj(h(2,:,:)), 2);
    
    % dekoder strumienia z anteny 1
    w1 = squeeze(W(1,:,:));
    y1 = sum(w1.*y);
    % dekoder strumienia z anteny 2
    w2 = squeeze(W(2,:,:));
    y2 = sum(w2.*y);
    
    % polacz strumienie transmitowane przez anten 1 i 2 w jeden strumien
    % odpowiadajacy strumieniowi oryginalnemu
    y_hat = zeros(1,N);
    y_hat(1:2:end) = y1;
    y_hat(2:2:end) = y2;
    
    % dekoder bpsk (kwantyzacja)
    bs_hat = real(y_hat)>0;
    
    % wyznacz BER
    nErr_MMSE(i) = sum(bs ~= bs_hat);
    fprintf('dekoder MMSE: BER= %.3d\n', nErr_ML(i)/N);
end


Ber_ML = nErr_ML/N;
Ber_MMSE = nErr_MMSE/N;

Eb_N0 = 10.^(Eb_N0_dB/10);
Ber_SISO = 0.5.*(1-1*(1+1./Eb_N0).^(-0.5));

p = 1/2 - 1/2*(1+1./Eb_N0).^(-1/2);
Ber_MRC_nRx2 = p.^2.*(1+2*(1-p));


figure;
title('Modulacja BPSK, kanal Rayleigha');
xlabel('Eb/No [dB]'); ylabel('BER');

semilogy(Eb_N0_dB, Ber_SISO, 'bp-', ...
         Eb_N0_dB, Ber_MMSE, 'ro-', ...
         Eb_N0_dB, Ber_ML,'mo-', ...
         Eb_N0_dB, Ber_MRC_nRx2, 'kd-', 'LineWidth',2);

axis([0 25 10^-5 0.5]); grid on
legend( 'SISO (nTx=1,nRx=1)', 
        'MMSE (nTx=2, nRx=2)',  
        'ML (nTx=2, nRx=2)', 
        'MRC (nTx=1,nRx=2)');
